

class Stream(object):
    """
    A very simple class that makes every file-like object a simple
    stream that only offers the read()-method.

     - **stream** the stream to wrap
    """

    def __init__(self, stream):
        self.stream = stream


    def read(self, n=-1):
        return self.stream.read(n)

    
        

class BufferedStream(object):


    def __init__(self, stream, bufsize=8192):
        self.stream = stream
        self.bufsize = bufsize
        self.pos = 0
        self.bufferoffset = 0
        self.buffer = ''


    def _read_from_stream(self, n=-1):
        data = self.stream.read(n)
        self.pos += len(data)
        return data
    

    def read(self, n=-1):
        # read everything. 
        if n == -1:
            res = ''
            if self.bufferoffset:
                res += self.buffer[self.bufferoffset:]
                self.bufferoffset = 0
            res += self._read_from_stream()
            self.buffer = res[-self.bufsize:]
            return res
        else:
            res = ''
            more = ''
            if self.bufferoffset:
                # this rather awkward way of slicing stems from the behavior of python
                # for string slicing.
                #
                # "abc"[-2:-1] == "b" # ok
                #
                # but
                #
                # "abc[-1:0] == "" # not what we want here.
                #
                # so we need to compute the to-index more explicit.
                res += self.buffer[self.bufferoffset:len(self.buffer) + self.bufferoffset + n]
                self.bufferoffset += n
                # we need to read something from the stream
                # to fulfill the request
                if self.bufferoffset > 0:
                    more = self._read_from_stream(self.bufferoffset)
                    self.bufferoffset = 0
            else:
                more = self._read_from_stream(n)

            # append + shift the buffer
            if more:
                res += more
                self.buffer = (self.buffer + more)[-self.bufsize:]

            return res
        

    def seek(self, position):
        assert position <= self.pos
        if position == self.pos:
            return
        assert self.pos - position <= len(self.buffer)
        self.bufferoffset = position - self.pos

        
    def tell(self):
        return self.pos + self.bufferoffset
    

    def __repr__(self):
        return "<%s pos: %i bufferoffset: %i >" % (self.__class__.__name__,
                                                self.pos,
                                                self.bufferoffset,
                                                )



from __future__ import with_statement
import tempfile

from unittest import TestCase

from abl.util import (
    Stream,
    BufferedStream,
    )



class TestStreams(TestCase):


    def test_simple_stream(self):
        tf = tempfile.mktemp()
        with open(tf, "w") as outf:
            outf.write("*" * 10)

        with open(tf) as inf:
            wf = Stream(inf)
            assert wf.read(5) == "*" * 5
            self.assertEqual(wf.read(), "*" * 5)


    def test_buffered_stream(self):
        testdata = "abcdefghijklmn" * 10000 # must be larger than our bufsize

        tf = tempfile.mktemp()
        with open(tf, "w") as outf:
            outf.write(testdata)

        with open(tf) as inf:
            wf = Stream(inf)
            bs = BufferedStream(wf)
            assert bs.read() == testdata

        with open(tf) as inf:
            wf = Stream(inf)
            bs = BufferedStream(wf)
            bs.read(100)
            bs.seek(0)
            assert bs.tell() == 0
            assert bs.read() == testdata

        # seek into the buffer,
        # then read a chunk over it's size
        with open(tf) as inf:
            wf = Stream(inf)
            bs = BufferedStream(wf)
            bs.read(100)
            bs.seek(10)
            assert bs.tell() == 10
            assert bs.read(300) == testdata[10:10 + 300]
            
            
        # seek into the buffer,
        # then read a chunk over it's size
        with open(tf) as inf:
            wf = Stream(inf)
            bs = BufferedStream(wf)
            bs.read(100)
            bs.seek(10)
            assert bs.tell() == 10
            assert bs.read(20) == testdata[10:10 + 20]
            assert bs.tell() == 30
            assert bs.read(1000) == testdata[30:1000 + 30]
            assert bs.tell() == 1030

        
        # read the last character out of the
        # buffer. That caused an error at some time.
        with open(tf) as inf:
            wf = Stream(inf)
            bs = BufferedStream(wf)
            bs.read(100)
            bs.seek(99)
            assert bs.read(1) == testdata[99]

        
            



from unittest import TestCase

import pickle

from abl.util import (
    Bunch,
    )

class Derived(Bunch):
    pass

class TestBunch(TestCase):

    def test_as_dict(self):
        bunch = Bunch(a='a', b='b')
        assert bunch == dict(a='a', b='b')

    def test_as_obj(self):
        bunch = Bunch(a='a', b='b')
        assert bunch.a == 'a'
        assert bunch.b == 'b'

    def test_failing_attribute(self):
        bunch = Bunch(a='a', b='b')
        self.assertRaises(AttributeError, getattr, bunch, 'c')

    def test_failing_key(self):
        bunch = Bunch(a='a', b='b')
        self.assertRaises(KeyError, lambda:bunch['c'])

    def test_pickling(self):
        bunch = Bunch(a='a', b='b')
        dump = pickle.dumps(bunch)
        from_pickle = pickle.loads(dump)
        assert bunch == from_pickle
        assert from_pickle.__class__ is Bunch

    def test_pickling_derived_class(self):

        derived = Derived(a='a', b='b')
        dump = pickle.dumps(derived)
        from_pickle = pickle.loads(dump)
        assert derived == from_pickle
        assert from_pickle.__class__ is Derived



import logging

def classproperty(f):
    """
    Create a property on a class, not an instance.

    Works only for getting.
    """
    class Descriptor(object):
        def __get__(self, _obj, objtype):
            return f(objtype)

    return Descriptor()


def partition(predicate, sequence):
    """
    Takes a predicate & a sequence of items, and
    applies the predicate to each of them.

    Returns a tuple with the first item being the
    list of matched items, and the second being
    the list of not matched items.
    """
    match, nomatch = [], []
    for item in sequence:
        if predicate(item):
            match.append(item)
        else:
            nomatch.append(item)
    return match, nomatch


class Bunch(dict):
    def __setattr__(self, key, item):
        self[key] = item


    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError


    def __delattr__(self, key):
        try:
            del self[key]
        except KeyError:
            raise AttributeError


    def __getstate__(self):
        return dict(**self)


    def __setstate__(self, data):
        self.update(data)


    def copy(self):
        return self.__class__(**super(Bunch, self).copy())


    def __repr__(self):
        res = ["<Bunch"]
        for name in self:
            if not name.startswith("__"):
                res.append("    '%s' = %r" % (name, self[name]))
        res.append("    >")
        return "\n".join(res)


#==============================================================================

class TeeBuffer(object):
    """
    Stream that buffers output and optionally also passes it to a given stream.
    """
    def __init__(self, oldstream=None):
        self.__buffer = []
        self._oldstream = oldstream


    def write(self, data):
        self.__buffer.append(data)
        if self._oldstream is not None:
            self._oldstream.write(data)


    def flush(self):
        if self._oldstream is not None:
            self._oldstream.flush()


    def getvalue(self):
        return "".join(self.__buffer)


def unicodify(text, codecs=('utf-8', 'latin-1', 'cp1252'), errors='ignore'):
    """
    cast any kind of string into a unicode string, even if the encoding is not known.
    If none of the codecs work, use 'unicode' with the errors attribute.
    """
    assert isinstance(text, basestring)
    if isinstance(text, unicode):
        return text
    for codec in codecs:
        try:
            return text.decode(codec)
        except UnicodeDecodeError:
            pass
    return unicode(text, errors=errors)


class NullHandler(logging.Handler):
    """
    The NullHandler can be used to set up logging so
    that no "no logger installed for ..." messages
    appear.

    Use it like this::

      logging.getLogger().addHandler(NullHandler())
      
    """
    
    def emit(self, *args, **kwargs):
        pass



from collections import defaultdict
from functools import partial

#==============================================================================

# A dict with scopes as keys and lists of memoized objects as values.
MEMOIZED = defaultdict(list)

def memoized(scope='global'):
    """
    Decorator that caches a function's return value when it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated.

     - param **scope**: You can hand in a scope to clean exactly the caches under
        this scope via calling `clear_full_scope` on the memoized function.
    """

    class Memoized(object):

        def __init__(self, func):
            self._func = func
            self._cache = {}
            self._scope = scope
            MEMOIZED[scope].append(self)


        def __call__(self, *args, **kwargs):
            key = args + tuple(kwargs.items())
            try:
                return self._cache[key]
            except KeyError:
                self._cache[key] = value = self._func(*args, **kwargs)
                return value
            except TypeError:
                # Uncachable -- for instance, passing a list or dict as an
                # argument. Better to not cache than to blow up entirely.
                return self._func(*args, **kwargs)


        def __get__(self, instance, class_):
            if instance is not None:
                return partial(self.__call__, instance)
            return self
                

        def __repr__(self):
            """Return the function's docstring."""
            return self._func.__doc__


        def clear_cache(self):
            """
            The cache that is used with this memoized decorators is cleaned.

            Usage:

                @memorized(SCOPE)
                def foo():
                    do_something()

                foo.clear_cache()
            """
            self._cache = {}


        def clear_full_scope(self):
            """
            All cached values that are used with memoized decorators with the
            same scope are cleaned.

            Usage:

                @memorized(SCOPE)
                def foo():
                    do_something()

                foo.clear_full_scope()
            """
            for memoized_obj in MEMOIZED[self._scope]:
                memoized_obj.clear_cache()


    if callable(scope):
        raise Exception('"memoized"-decorator was not called on initialization.')

    return Memoized




from unittest import TestCase

from abl.util import memoized


class MemoizationTests(TestCase):


    def test_memoization_works(self):
        global COUNTER
        COUNTER = 0
        def counter():
            global COUNTER
            COUNTER += 1
            return COUNTER

        # first, make sure the function
        # works.
        assert counter() == 1
        # now memoize it, and
        # subsequent calls will return the
        # same value.
        counter = memoized()(counter)
        assert counter() == 2
        assert counter() == 2


    def test_memoization_fails_gracefully_on_non_hashable_arguments(self):
        global COUNTER
        COUNTER = 0
        
        @memoized()
        def test(arg):
            global COUNTER
            COUNTER += 1
            return str(arg)


        assert test(10) == test(10) and COUNTER == 1

        assert test({}) == test({}) and COUNTER == 3


    def test_memoization_doesnt_mask_exceptions(self):
        global COUNTER
        COUNTER = 0
        
        @memoized()
        def test():
            global COUNTER
            COUNTER += 1
            raise TypeError

        self.failUnlessRaises(TypeError, test)
        # we are only called *once*
        assert COUNTER == 1
        

        COUNTER = 0
        
        @memoized()
        def test():
            global COUNTER
            COUNTER += 1
            raise KeyError

        self.failUnlessRaises(KeyError, test)
        # we are only called *once*
        assert COUNTER == 1
        

    def test_clearing_memoization_cache(self):
        global COUNTER
        COUNTER = 0

        @memoized()
        def test():
            global COUNTER
            COUNTER += 1
            return COUNTER

        @memoized()
        def test_two():
            global COUNTER
            COUNTER += 1
            return COUNTER

        assert test() == 1
        assert test_two() == 2
        test.clear_cache()
        # only the test-cache must be cleared.
        assert test() == 3
        assert test_two() == 2

        # now clear the full global scope.
        test.clear_full_scope()
        assert test() == 4
        assert test_two() == 5
        assert test() == 4
        assert test_two() == 5
        
        
        
    def test_memoization_on_objects(self):
        global COUNTER
        COUNTER = 0
        
        class Test(object):


            @memoized()
            def test_method(self):
                global COUNTER
                COUNTER += 1
                return COUNTER

            @classmethod
            @memoized()
            def test_classmethod(cls):
                global COUNTER
                COUNTER += 1
                return COUNTER


        a = Test()
        b = Test()

        assert a.test_method() == 1
        assert a.test_method() == 1
        assert b.test_method() == 2
        assert b.test_method() == 2
        assert a.test_classmethod() == 3
        assert b.test_classmethod() == 3
        

    def test_scope_isolation(self):
        global COUNTER
        COUNTER = 0
        
        @memoized("foo")
        def test_foo():
            global COUNTER
            COUNTER += 1
            return COUNTER

        @memoized("bar")
        def test_bar():
            global COUNTER
            COUNTER += 1
            return COUNTER

        assert test_foo() == 1
        assert test_bar() == 2
        assert test_foo() == 1
        assert test_bar() == 2

        test_foo.clear_full_scope()
        assert test_foo() == 3
        assert test_bar() == 2



from __future__ import absolute_import

from .stream import (
    Stream,
    BufferedStream,
    )

from .memoization import (
    memoized,
    )

from .misc import (
    unicodify,
    classproperty,
    partition,
    Bunch,
    TeeBuffer,
    NullHandler,
    )



# -*- coding: utf-8 -*-
from unittest import TestCase

from abl.util import (
    classproperty,
    partition,
    unicodify,
    )

class TestMisc(TestCase):


    def test_classproperty(self):

        value = 100
        class Foo(object):

            FOO = value
            
            @classproperty
            def test(cls):
                return cls.FOO

        assert Foo.test == value
        assert Foo().test == value


    def test_partition(self):
        self.assertEqual(partition(lambda i: i < 5, xrange(10)),
                         ([0, 1, 2, 3, 4], [5, 6, 7, 8, 9]))

class TestUnicodify(TestCase):

    def test_unicode_unity(self):
        utest = u"Hülle"
        assert unicodify(utest) is utest

    def test_utf8_to_unicode(self):
        teststring = 'H\xc3\xbclle'
        assert unicodify(teststring) == u"Hülle"

    def test_ignore_errors(self):
        teststring = 'H\xc3\xbclle'
        assert unicodify(teststring, codecs=()) == u"Hlle"




from setuptools import setup, find_packages


TEST_REQUIREMENTS = ["nose"]

setup(
    name = "abl.util",
    version = "0.1.1",
    author = "Diez B. Roggisch",
    author_email = "diez.roggisch@ableton.com",
    description = "A package that contains various helpful classes and functions used widely in the Ableton Python code base.",
    license="MIT",
    packages = find_packages(),
    namespace_packages = ["abl"],
    install_requires = [
        ],
    extras_require = dict(
        testing=TEST_REQUIREMENTS,
        ),
    tests_require=TEST_REQUIREMENTS,
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Topic :: Utilities',
    ],
    zip_safe=False,
)

